import loam.tools


def test_set_conf_str(conf):
    loam.tools.set_conf_str(conf, ['sectionB.optA=42', 'sectionA.optBool=f'])
    assert conf.sectionB.optA == 42
    assert conf.sectionA.optBool is False



import pytest
import loam.error
import loam.manager


def test_build_from_dict(conf, conf_def):
    conf_dict = {name: dict(sct.def_) for name, sct in conf_def.items()}
    conf_fd = loam.manager.ConfigurationManager.from_dict_(conf_dict)
    assert all(conf[s][o] == conf_fd[s][o]
               for s in conf_def for o in conf_def[s])


def test_build_section_invalid_option_name():
    invalid_id = 'not a valid id'
    with pytest.raises(loam.error.OptionError) as err:
        loam.manager.Section(**{invalid_id: 'dummy'})
    assert err.value.option == invalid_id


def test_build_manager_invalid_section_name():
    invalid_id = 'not a valid id'
    with pytest.raises(loam.error.SectionError) as err:
        loam.manager.ConfigurationManager(**{invalid_id: 'dummy'})
    assert err.value.section == invalid_id


def test_get_subconfig(conf, conf_def):
    for sub in conf_def:
        assert getattr(conf, sub) is conf[sub]


def test_get_opt(conf):
    for sub, opt in conf.options_():
        assert getattr(conf[sub], opt) is conf[sub][opt]


def test_set_opt_def(conf):
    with pytest.raises(TypeError):
        conf.sectionA.def_['optA'] = None


def test_get_invalid_subconfig(conf):
    invalid = 'invalidsubdummy'
    with pytest.raises(loam.error.SectionError) as err:
        _ = conf[invalid]
    assert err.value.section == invalid


def test_get_invalid_opt(conf):
    invalid = 'invalidoptdummy'
    for sub in conf:
        with pytest.raises(loam.error.OptionError) as err:
            _ = conf[sub][invalid]
        assert err.value.option == invalid


def test_contains_section(conf, conf_def):
    for sub in conf_def:
        assert sub in conf


def test_contains_option(conf, conf_def):
    for sub, opts in conf_def.items():
        for opt in opts:
            assert opt in conf[sub]


def test_contains_invalid_section(conf):
    assert 'invalidsubdummy' not in conf


def test_contains_invalid_option(conf):
    assert 'invalidoptdummy' not in conf.sectionA


def test_reset_all(conf):
    conf.sectionA.optA = 42
    conf.reset_()
    assert conf.sectionA.optA == 1


def test_reset_subconfig(conf):
    conf.sectionA.optA = 42
    del conf.sectionA
    assert conf.sectionA.optA == 1


def test_reset_subconfig_item(conf):
    conf.sectionA.optA = 42
    del conf['sectionA']
    assert conf.sectionA.optA == 1


def test_reset_subconfig_func(conf):
    conf.sectionA.optA = 42
    conf.sectionA.reset_()
    assert conf.sectionA.optA == 1


def test_reset_opt(conf):
    conf.sectionA.optA = 42
    del conf.sectionA.optA
    assert conf.sectionA.optA == 1


def test_reset_opt_item(conf):
    conf.sectionA.optA = 42
    del conf.sectionA['optA']
    assert conf.sectionA.optA == 1


def test_update_opt(conf):
    conf.sectionA.update_({'optA': 42, 'optC': 43})
    assert conf.sectionA.optA == 42 and conf.sectionA.optC == 43


def test_update_section(conf):
    conf.update_({'sectionA': {'optA': 42}, 'sectionB': {'optA': 43}})
    assert conf.sectionA.optA == 42 and conf.sectionB.optA == 43


def test_update_opt_conf_arg(conf):
    conf.sectionA.update_({'optB': 42})
    assert conf.sectionA.optB == 2
    conf.sectionA.update_({'optB': 42}, conf_arg=False)
    assert conf.sectionA.optB == 42


def test_update_section_conf_arg(conf):
    conf.update_({'sectionA': {'optB': 42}, 'sectionB': {'optB': 43}})
    assert conf.sectionA.optB == 2 and conf.sectionB.optB == 5
    conf.update_({'sectionA': {'optB': 42}, 'sectionB': {'optB': 43}},
                 conf_arg=False)
    assert conf.sectionA.optB == 42 and conf.sectionB.optB == 43


def test_opt_def_values(conf, conf_def):
    assert all(conf[s].def_[o] == conf_def[s].def_[o]
               for s in conf_def for o in conf_def[s])


def test_config_iter_subs(conf, conf_def):
    raw_iter = set(iter(conf))
    scts_iter = set(conf.sections_())
    scts_expected = set(conf_def.keys())
    assert raw_iter == scts_iter == scts_expected


def test_config_iter_options(conf, conf_def):
    options_iter = set(conf.options_())
    options_expected = set((sub, opt) for sub in conf_def
                           for opt in conf_def[sub])
    assert options_iter == options_expected


def test_config_iter_default_val(conf):
    vals_iter = set(conf.opt_vals_())
    vals_dflts = set((s, o, m.default) for s, o, m in conf.defaults_())
    assert vals_iter == vals_dflts


def test_config_iter_subconfig(conf, conf_def):
    raw_iter = set(iter(conf.sectionA))
    opts_iter = set(conf.sectionA.options_())
    opts_expected = set(conf_def['sectionA'])
    assert raw_iter == opts_iter == opts_expected


def test_config_iter_subconfig_default_val(conf):
    vals_iter = set(conf.sectionA.opt_vals_())
    vals_dflts = set((o, m.default) for o, m in conf.sectionA.defaults_())
    assert vals_iter == vals_dflts


def test_context_section(conf):
    with conf.sectionA.context_(optA=42, optBool=False):
        assert conf.sectionA.optA == 42
        assert conf.sectionA.optBool is False
    assert conf.sectionA.optA == 1
    assert conf.sectionA.optBool is True



"""Definition of configuration manager classes.

Note:
    All methods and attributes are postfixed with an underscore to minimize the
    risk of collision with the names of your configuration sections and
    options.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from types import MappingProxyType
import typing

import toml

from . import error, _internal

if typing.TYPE_CHECKING:
    from typing import (Dict, List, Any, Union, Tuple, Mapping, Optional,
                        Iterator, Iterable, ContextManager)
    from os import PathLike


def _is_valid(name: str) -> bool:
    """Check if a section or option name is valid."""
    return name.isidentifier() and name != 'loam_sub_name'


@dataclass(frozen=True)
class ConfOpt:
    """Metadata of configuration options.

    Attributes:
        default: the default value of the configuration option.
        cmd_arg: whether the option is a command line argument.
        shortname: short version of the command line argument.
        cmd_kwargs: keyword arguments fed to
            :meth:`argparse.ArgumentParser.add_argument` during the
            construction of the command line arguments parser.
        conf_arg: whether the option can be set in the config file.
        help: short description of the option.
        comprule: completion rule for ZSH shell.
    """

    default: Any
    cmd_arg: bool = False
    shortname: Optional[str] = None
    cmd_kwargs: Dict[str, Any] = field(default_factory=dict)
    conf_arg: bool = False
    help: str = ''
    comprule: Optional[str] = ''


class Section:
    """Hold options for a single section.

    Args:
        options: option metadata. The name of each *option* is the name of the
            keyword argument passed on to this function. Option names should be
            valid identifiers, otherwise an :class:`~loam.error.OptionError` is
            raised.
    """

    def __init__(self, **options: ConfOpt):
        self._def = {}
        for opt_name, opt_meta in options.items():
            if _is_valid(opt_name):
                self._def[opt_name] = opt_meta
                self[opt_name] = opt_meta.default
            else:
                raise error.OptionError(opt_name)

    @property
    def def_(self) -> Mapping[str, Any]:
        """Return mapping of default values of options."""
        return MappingProxyType(self._def)

    def __getitem__(self, opt: str) -> Any:
        return getattr(self, opt)

    def __setitem__(self, opt: str, value: Any):
        setattr(self, opt, value)

    def __delitem__(self, opt: str):
        delattr(self, opt)

    def __delattr__(self, opt: str):
        if opt not in self:
            raise error.OptionError(opt)
        self[opt] = self.def_[opt].default

    def __getattr__(self, opt: str):
        raise error.OptionError(opt)

    def __iter__(self) -> Iterator[str]:
        return iter(self.def_.keys())

    def __contains__(self, opt: str) -> bool:
        return opt in self.def_

    def options_(self) -> Iterator[str]:
        """Iterate over configuration option names."""
        return iter(self)

    def opt_vals_(self) -> Iterator[Tuple[str, Any]]:
        """Iterate over option names and option values."""
        for opt in self.options_():
            yield opt, self[opt]

    def defaults_(self) -> Iterable[Tuple[str, ConfOpt]]:
        """Iterate over option names, and option metadata."""
        return self.def_.items()

    def update_(self, sct_dict: Mapping[str, Any], conf_arg: bool = True):
        """Update values of configuration section with dict.

        Args:
            sct_dict: mapping of option names to their new values. Unknown
                options are discarded.
            conf_arg: if True, only options that can be set in a config file
                are updated.
        """
        for opt, val in sct_dict.items():
            if opt not in self.def_:
                continue
            if not conf_arg or self.def_[opt].conf_arg:
                self[opt] = val

    def reset_(self):
        """Restore default values of options in this section."""
        for opt, meta in self.defaults_():
            self[opt] = meta.default

    def context_(self, **options: Any) -> ContextManager[None]:
        """Enter a context to locally change option values.

        This context is reusable but not reentrant.
        """
        return _internal.SectionContext(self, options)


class ConfigurationManager:
    """Configuration manager.

    Configuration options are organized in sections. A configuration option can
    be accessed both with attribute and item access notations, these two lines
    access the same option value::

        conf.some_section.some_option
        conf['some_section']['some_option']

    To reset a configuration option (or an entire section) to its default
    value, simply delete it (with item or attribute notation)::

        del conf['some_section']  # reset all options in 'some_section'
        del conf.some_section.some_option  # reset a particular option

    It will be set to its default value the next time you access it.

    Args:
        sections: section metadata. The name of each *section* is the name of
            the keyword argument passed on to this function. Section names
            should be valid identifiers, otherwise a
            :class:`~loam.error.SectionError` is raised.
    """

    def __init__(self, **sections: Section):
        self._sections = []
        for sct_name, sct_meta in sections.items():
            if _is_valid(sct_name):
                setattr(self, sct_name, Section(**sct_meta.def_))
                self._sections.append(sct_name)
            else:
                raise error.SectionError(sct_name)
        self._parser = None
        self._nosub_valid = False
        self._config_files: Tuple[Path, ...] = ()

    @classmethod
    def from_dict_(
        cls, conf_dict: Mapping[str, Mapping[str, ConfOpt]]
    ) -> ConfigurationManager:
        """Use a dictionary to create a :class:`ConfigurationManager`.

        Args:
            conf_dict: the first level of keys are section names. The second
                level are option names. The values are the options metadata.

        Returns:
            a configuration manager with the requested sections and options.
        """
        return cls(**{name: Section(**opts)
                      for name, opts in conf_dict.items()})

    @property
    def config_files_(self) -> Tuple[Path, ...]:
        """Path of config files.

        The config files are in the order of reading. This means the most
        global config file is the first one on this list while the most local
        config file is the last one.
        """
        return self._config_files

    def set_config_files_(self, *config_files: Union[str, PathLike]):
        """Set the list of config files.

        Args:
            config_files: path of config files, given in the order
                of reading.
        """
        self._config_files = tuple(Path(path) for path in config_files)

    def __getitem__(self, sct: str) -> Section:
        return getattr(self, sct)

    def __delitem__(self, sct: str):
        delattr(self, sct)

    def __delattr__(self, sct: str):
        self[sct].reset_()

    def __getattr__(self, sct: str):
        raise error.SectionError(sct)

    def __iter__(self) -> Iterator[str]:
        return iter(self._sections)

    def __contains__(self, sct: str) -> bool:
        return sct in self._sections

    def sections_(self) -> Iterator[str]:
        """Iterate over configuration section names."""
        return iter(self)

    def options_(self) -> Iterator[Tuple[str, str]]:
        """Iterate over section and option names.

        This iterator is also implemented at the section level. The two loops
        produce the same output::

            for sct, opt in conf.options_():
                print(sct, opt)

            for sct in conf.sections_():
                for opt in conf[sct].options_():
                    print(sct, opt)
        """
        for sct in self:
            for opt in self[sct]:
                yield sct, opt

    def opt_vals_(self) -> Iterator[Tuple[str, str, Any]]:
        """Iterate over sections, option names, and option values.

        This iterator is also implemented at the section level. The two loops
        produce the same output::

            for sct, opt, val in conf.opt_vals_():
                print(sct, opt, val)

            for sct in conf.sections_():
                for opt, val in conf[sct].opt_vals_():
                    print(sct, opt, val)
        """
        for sct, opt in self.options_():
            yield sct, opt, self[sct][opt]

    def defaults_(self) -> Iterator[Tuple[str, str, Any]]:
        """Iterate over sections, option names, and option metadata.

        This iterator is also implemented at the section level. The two loops
        produce the same output::

            for sct, opt, meta in conf.defaults_():
                print(sct, opt, meta.default)

            for sct in conf.sections_():
                for opt, meta in conf[sct].defaults_():
                    print(sct, opt, meta.default)
        """
        for sct, opt in self.options_():
            yield sct, opt, self[sct].def_[opt]

    def reset_(self):
        """Restore default values of all options."""
        for sct, opt, meta in self.defaults_():
            self[sct][opt] = meta.default

    def create_config_(self, index: int = 0, update: bool = False):
        """Create config file.

        Create config file in :attr:`config_files_[index]`.

        Parameters:
            index: index of config file.
            update: if set to True and :attr:`config_files_` already exists,
                its content is read and all the options it sets are kept in the
                produced config file.
        """
        if not self.config_files_[index:]:
            return
        path = self.config_files_[index]
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        conf_dict: Dict[str, Dict[str, Any]] = {}
        for section in self.sections_():
            conf_opts = [o for o, m in self[section].defaults_() if m.conf_arg]
            if not conf_opts:
                continue
            conf_dict[section] = {}
            for opt in conf_opts:
                conf_dict[section][opt] = (self[section][opt] if update else
                                           self[section].def_[opt].default)
        with path.open('w') as cfile:
            toml.dump(conf_dict, cfile)

    def update_(self, conf_dict: Mapping[str, Mapping[str, Any]],
                conf_arg: bool = True):
        """Update values of configuration options with dict.

        Args:
            conf_dict: new values indexed by section and option names.
            conf_arg: if True, only options that can be set in a config file
                are updated.
        """
        for section, secdict in conf_dict.items():
            self[section].update_(secdict, conf_arg)

    def read_config_(self, cfile: Path) -> Optional[Mapping]:
        """Read a config file and set config values accordingly.

        Returns:
            content of config file.
        """
        if not cfile.exists():
            return {}
        try:
            conf_dict = toml.load(str(cfile))
        except toml.TomlDecodeError:
            return None
        self.update_(conf_dict)
        return conf_dict

    def read_configs_(self) -> Tuple[Dict[str, Dict[str, Any]], List[Path],
                                     List[Path]]:
        """Read config files and set config values accordingly.

        Returns:
            respectively content of files, list of missing/empty files and list
            of files for which a parsing error arised.
        """
        if not self.config_files_:
            return {}, [], []
        content: Dict[str, Dict[str, Any]] = {section: {} for section in self}
        empty_files = []
        faulty_files = []
        for cfile in self.config_files_:
            conf_dict = self.read_config_(cfile)
            if conf_dict is None:
                faulty_files.append(cfile)
                continue
            elif not conf_dict:
                empty_files.append(cfile)
                continue
            for section, secdict in conf_dict.items():
                content[section].update(secdict)
        return content, empty_files, faulty_files



import pytest

from loam.manager import ConfOpt, Section, ConfigurationManager
from loam.cli import Subcmd, CLIManager
from loam.tools import switch_opt


@pytest.fixture(scope='session', params=['confA'])
def conf_def(request):
    metas = {}
    metas['confA'] = {
        'sectionA': Section(
            optA=ConfOpt(1, True, 'a', {}, True, 'AA'),
            optB=ConfOpt(2, True, None, {}, False, 'AB'),
            optC=ConfOpt(3, True, None, {}, True, 'AC'),
            optBool=switch_opt(True, 'o', 'Abool'),
        ),
        'sectionB': Section(
            optA=ConfOpt(4, True, None, {}, True, 'BA'),
            optB=ConfOpt(5, True, None, {}, False, 'BB'),
            optC=ConfOpt(6, False, None, {}, True, 'BC'),
            optBool=switch_opt(False, 'o', 'Bbool'),
        ),
    }
    return metas[request.param]


@pytest.fixture
def conf(conf_def):
    return ConfigurationManager(**conf_def)


@pytest.fixture(params=['subsA'])
def sub_cmds(request):
    subs = {}
    subs['subsA'] = {
        'common_': Subcmd('subsA loam test'),
        'bare_': Subcmd(None, 'sectionA'),
        'sectionB': Subcmd('sectionB subcmd help'),
    }
    return subs[request.param]


@pytest.fixture
def climan(conf, sub_cmds):
    return CLIManager(conf, **sub_cmds)


@pytest.fixture
def cfile(tmp_path):
    return tmp_path / 'config.toml'


@pytest.fixture
def nonexistent_file(tmp_path):
    return tmp_path / 'dummy.toml'


@pytest.fixture
def illtoml(tmp_path):
    path = tmp_path / 'ill.toml'
    path.write_text('not}valid[toml\n')
    return path



import toml


def test_config_files_setter(conf, cfile):
    assert conf.config_files_ == ()
    conf.set_config_files_(str(cfile))
    assert conf.config_files_ == (cfile,)


def test_create_config(conf, cfile):
    conf.set_config_files_(cfile)
    conf.create_config_()
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_create_config_index(conf, cfile, nonexistent_file):
    conf.set_config_files_(nonexistent_file, cfile)
    conf.create_config_(index=1)
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_create_config_no_update(conf, cfile):
    conf.set_config_files_(cfile)
    conf.sectionA.optA = 42
    conf.create_config_()
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_create_config_update(conf, cfile):
    conf.set_config_files_(cfile)
    conf.sectionA.optA = 42
    conf.create_config_(update=True)
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 42, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_read_config(conf, cfile):
    conf.set_config_files_(cfile)
    conf.create_config_()
    conf_dict = conf.read_config_(cfile)
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_read_config_missing(conf, cfile):
    assert conf.read_config_(cfile) == {}


def test_read_config_empty(conf, cfile):
    cfile.touch()
    assert conf.read_config_(cfile) == {}


def test_read_config_invalid(conf, illtoml):
    assert conf.read_config_(illtoml) is None


def test_read_configs(conf, cfile):
    conf.set_config_files_(cfile)
    conf.create_config_()
    conf_dict, empty, faulty = conf.read_configs_()
    assert empty == faulty == []
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_read_configs_missing_invalid(conf, nonexistent_file, illtoml):
    conf.set_config_files_(illtoml, nonexistent_file)
    conf_dict, empty, faulty = conf.read_configs_()
    assert empty == [nonexistent_file]
    assert faulty == [illtoml]
    assert conf_dict == {'sectionA': {}, 'sectionB': {}}



"""Various helper functions and classes.

They are designed to help you use :class:`~loam.manager.ConfigurationManager`.
"""

from __future__ import annotations
import pathlib
import subprocess
import shlex
import typing

from . import error, _internal
from .manager import ConfOpt

if typing.TYPE_CHECKING:
    from typing import Optional, Dict, List, Union
    from os import PathLike
    from .manager import ConfigurationManager
    from .cli import CLIManager


def switch_opt(default: bool, shortname: Optional[str],
               help_msg: str) -> ConfOpt:
    """Define a switchable ConfOpt.

    This creates a boolean option. If you use it in your CLI, it can be
    switched on and off by prepending + or - to its name: +opt / -opt.

    Args:
        default: the default value of the swith option.
        shortname: short name of the option, no shortname will be used if set
            to None.
        help_msg: short description of the option.

    Returns:
        a :class:`~loam.manager.ConfOpt` with the relevant properties.
    """
    return ConfOpt(bool(default), True, shortname,
                   dict(action=_internal.Switch), True, help_msg, None)


def command_flag(shortname: Optional[str], help_msg: str) -> ConfOpt:
    """Define a command line flag.

    The corresponding option is set to true if it is passed as a command line
    option.  This is similar to :func:`switch_opt`, except the option is not
    available from config files.  There is therefore no need for a mechanism to
    switch it off from the command line.

    Args:
        shortname: short name of the option, no shortname will be used if set
            to None.
        help_msg: short description of the option.

    Returns:
        a :class:`~loam.manager.ConfOpt` with the relevant properties.
    """
    return ConfOpt(None, True, shortname, dict(action='store_true'), False,
                   help_msg, None)


def config_conf_section() -> Dict[str, ConfOpt]:
    """Define a configuration section handling config file.

    Returns:
        definition of the 'create', 'create_local', 'update', 'edit' and
        'editor' configuration options.
    """
    return dict(
        create=command_flag(None, 'create most global config file'),
        create_local=command_flag(None, 'create most local config file'),
        update=command_flag(None, 'add missing entries to config file'),
        edit=command_flag(None, 'open config file in a text editor'),
        editor=ConfOpt('vim', conf_arg=True, help='text editor'),
    )


def set_conf_opt(shortname: Optional[str] = None) -> ConfOpt:
    """Define a Confopt to set a config option.

    You can feed the value of this option to :func:`set_conf_str`.

    Args:
        shortname: shortname for the option if relevant.

    Returns:
        the option definition.
    """
    return ConfOpt(None, True, shortname,
                   dict(action='append', metavar='section.option=value'),
                   False, 'set configuration options')


def set_conf_str(conf: ConfigurationManager, optstrs: List[str]):
    """Set options from a list of section.option=value string.

    Args:
        conf: the :class:`~loam.manager.ConfigurationManager` to update.
        optstrs: the list of 'section.option=value' formatted strings.
    """
    falsy = ['0', 'no', 'n', 'off', 'false', 'f']
    bool_actions = ['store_true', 'store_false', _internal.Switch]
    for optstr in optstrs:
        opt, val = optstr.split('=', 1)
        sec, opt = opt.split('.', 1)
        if sec not in conf:
            raise error.SectionError(sec)
        if opt not in conf[sec]:
            raise error.OptionError(opt)
        meta = conf[sec].def_[opt]
        if meta.default is None:
            if 'type' in meta.cmd_kwargs:
                cast = meta.cmd_kwargs['type']
            else:
                act = meta.cmd_kwargs.get('action')
                cast = bool if act in bool_actions else str
        else:
            cast = type(meta.default)
        if cast is bool and val.lower() in falsy:
            val = ''
        conf[sec][opt] = cast(val)


def config_cmd_handler(conf: ConfigurationManager, config: str = 'config'):
    """Implement the behavior of a subcmd using config_conf_section.

    Args:
        conf: a :class:`~loam.manager.ConfigurationManager` containing a
            section created with :func:`config_conf_section` function.
        config: name of the configuration section created with
            :func:`config_conf_section` function.
    """
    if conf[config].create or conf[config].update:
        conf.create_config_(update=conf[config].update)
    if conf[config].create_local:
        conf.create_config_(index=-1, update=conf[config].update)
    if conf[config].edit:
        if not conf.config_files_[0].is_file():
            conf.create_config_(update=conf[config].update)
        subprocess.run(shlex.split('{} {}'.format(conf[config].editor,
                                                  conf.config_files_[0])))


def create_complete_files(climan: CLIManager, path: Union[str, PathLike],
                          cmd: str, *cmds: str, zsh_sourceable: bool = False,
                          zsh_force_grouping: bool = False):
    """Create completion files for bash and zsh.

    Args:
        climan: a :class:`~loam.cli.CLIManager`.
        path: directory in which the config files should be created. It is
            created if it doesn't exist.
        cmd: command name that should be completed.
        cmds: extra command names that should be completed.
        zsh_sourceable: if True, the generated file will contain an explicit
            call to ``compdef``, which means it can be sourced to activate CLI
            completion.
        zsh_force_grouping: if True, assume zsh supports grouping of options.
            Otherwise, loam will attempt to check whether zsh >= 5.4.
    """
    path = pathlib.Path(path)
    zsh_dir = path / 'zsh'
    zsh_dir.mkdir(parents=True, exist_ok=True)
    zsh_file = zsh_dir / f"_{cmd}.sh"
    bash_dir = path / 'bash'
    bash_dir.mkdir(parents=True, exist_ok=True)
    bash_file = bash_dir / f"{cmd}.sh"
    climan.zsh_complete(zsh_file, cmd, *cmds, sourceable=zsh_sourceable,
                        force_grouping=zsh_force_grouping)
    climan.bash_complete(bash_file, cmd, *cmds)



from pathlib import Path
from setuptools import setup


setup(
    name='loam',

    description='Light configuration manager',
    long_description=Path("README.rst").read_text(),

    url='https://github.com/amorison/loam',

    author='Adrien Morison',
    author_email='adrien.morison@gmail.com',

    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    packages=['loam'],
    package_data={'loam': ['py.typed']},

    python_requires=">=3.7",
    install_requires=[
        'setuptools_scm>=6.2',
        'toml>=0.10.2',
    ],
)
